/**
 * Obfuscator sample
 * (c) Copyright IBM 2009
 */
package com.ibm.rhapsody.samples.obfuscator;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Random;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;

import com.telelogic.rhapsody.core.IRPAction;
import com.telelogic.rhapsody.core.IRPAnnotation;
import com.telelogic.rhapsody.core.IRPApplication;
import com.telelogic.rhapsody.core.IRPClass;
import com.telelogic.rhapsody.core.IRPCollection;
import com.telelogic.rhapsody.core.IRPDiagram;
import com.telelogic.rhapsody.core.IRPEventReception;
import com.telelogic.rhapsody.core.IRPGraphElement;
import com.telelogic.rhapsody.core.IRPGraphicalProperty;
import com.telelogic.rhapsody.core.IRPGuard;
import com.telelogic.rhapsody.core.IRPInstance;
import com.telelogic.rhapsody.core.IRPMessage;
import com.telelogic.rhapsody.core.IRPModelElement;
import com.telelogic.rhapsody.core.IRPOperation;
import com.telelogic.rhapsody.core.IRPPackage;
import com.telelogic.rhapsody.core.IRPProject;
import com.telelogic.rhapsody.core.IRPRelation;
import com.telelogic.rhapsody.core.IRPState;
import com.telelogic.rhapsody.core.IRPStereotype;
import com.telelogic.rhapsody.core.IRPTransition;
import com.telelogic.rhapsody.core.IRPType;
import com.telelogic.rhapsody.core.IRPUnit;
import com.telelogic.rhapsody.core.IRPVariable;
import com.telelogic.rhapsody.core.RPUserPlugin;
import com.telelogic.rhapsody.core.RhapsodyAppServer;
import com.telelogic.rhapsody.core.RhapsodyRuntimeException;

public class RhpObfuscator extends RPUserPlugin {
	
	static class Obfuscator {
		
		private static class DealWithUnit {
			boolean result;
			boolean askAgain = true;
		}
		
		private static class ProgressUpdater extends Thread {
			
			JLabel progressLabel;
			
			JTextField[] progressLabelPercent = new JTextField[100];
			JFrame progressFrame;
			
	
			ProgressUpdater() {
				
				JPanel panel = new JPanel();
				panel.setPreferredSize(new Dimension(400, 100));
				panel.setBounds(30, 30, 200, 200);
				panel.setLayout(new GridLayout(1,1));
				JPanel progressPanel = new JPanel();
				JPanel progressPanelRestricter = new JPanel();
				progressPanelRestricter.setLayout(new GridLayout(3,1));
				progressPanelRestricter.add(new JLabel());
				progressPanelRestricter.add(progressPanel);
				progressPanel.setLayout(new GridLayout(1,100));
				for (int i = 0; i < progressLabelPercent.length; i++) {
					progressLabelPercent[i] = new JTextField(" ");
					progressLabelPercent[i].setEditable(false);
					progressLabelPercent[i].setBorder(null);
					progressLabelPercent[i].setBackground(Color.black);
					progressLabelPercent[i].setForeground(Color.black);
					progressLabelPercent[i].setPreferredSize(new Dimension(2,10));
					progressLabelPercent[i].setMaximumSize(new Dimension(100,10));
					progressPanel.add(progressLabelPercent[i]);
				}
				progressLabel = new JLabel("  Obfuscation Progress: " + "0" + "%");
				progressLabel.setPreferredSize(new Dimension(200, 20));
				panel.add(progressLabel);
				progressFrame = new JFrame();
				progressFrame.setLayout(new GridLayout(2,1));
				progressFrame.setPreferredSize(new Dimension(400, 200));
				progressFrame.add(panel);
				progressFrame.add(progressPanelRestricter);
				progressFrame.setVisible(true);
				panel.repaint();
				Dimension screenSize = java.awt.Toolkit.getDefaultToolkit().getScreenSize();
				progressFrame.setBounds(screenSize.width/2-300/2, screenSize.height/2-200/2, 300, 200);
				
			}

			public void setProgress(int progress) {
				progressLabel.setText("Progress: " + progress + "%");
				progressLabel.repaint();
				for (int i = 0; i < progress; i++) {
					progressLabelPercent[i].setBackground(Color.GREEN);
					progressLabelPercent[i].setForeground(Color.GREEN);
				}
			}
			
			public void bringToFront() {
				if (progressFrame != null) {
					progressFrame.toFront();
				}
			}

			public void close() {
				if (progressFrame != null) {
					progressFrame.dispose();
				}
				
			}
			
		}
		
		private static final Integer aAscii = 64;

		private static boolean bStripFilename = true;

		private static Map<String, String> obfuscatedNames;

		private static Random r = new Random(System.currentTimeMillis());
		
		private static DealWithUnit unloaded = new DealWithUnit();
		private static DealWithUnit reference = new DealWithUnit();
		private static DealWithUnit notCopied = new DealWithUnit();
		private static DealWithUnit readOnly = new DealWithUnit();

		private static List<String> predefinedTypesPackages = Arrays.asList(
				new String[] {
						"PredefinedTypes".toLowerCase(),
						"PredefinedTypesAda".toLowerCase(),
						"PredefinedTypesCSharp".toLowerCase(),
						"PredefinedTypesCpp".toLowerCase(),
						"PredefinedTypesC".toLowerCase(),
						"PredefinedTypesJava".toLowerCase(),
						"PredefinedTypesRAL".toLowerCase()
				}
				);
		
		private static List<String> unitsToSetPath = new ArrayList<String>();
		private static List<String> unitsToDelete = new ArrayList<String>();
		
		public static List<String> unitsNotToObfuscate;

		public static List<String> propertiesToObfuscate;

		private static Properties obfuscatorProperties;

		private static IRPModelElement rootElement;

		private static IRPProject rootProject;

		// generates random name for given name - same length
		private static String GetNewName(String oldName) {

			if (oldName.length() == 0)
				return "";

			if (obfuscatedNames == null)
				obfuscatedNames = new HashMap<String, String>();

			String newName = obfuscatedNames.get(oldName);
			if (newName == null) {

				newName = "";
				char[] oldNameCharArray = oldName.toCharArray();
				for (int i = 0; i < oldName.length(); i++) {
					if (oldNameCharArray[i] == ' ')
						newName += ' ';
					else
						newName += (char) (aAscii + (int) ((26 * r.nextFloat()) + 1));

				}
				obfuscatedNames.put(oldName, newName);
			}
			return newName;
		}

		// ask user if wants to obfuscate unit file names
		private static boolean initShouldObfuscateUnitFileNames() {
			obfuscatedNames = null;
			String strMessage = "Do you want to obfuscate the unit filenames as well?";
			int nAnswer = JOptionPane.showOptionDialog(null, strMessage,
					"Obfuscator", JOptionPane.YES_NO_CANCEL_OPTION,
					JOptionPane.QUESTION_MESSAGE, null, null, null);
			if (nAnswer == 0) // yes
				bStripFilename = true;
			else if (nAnswer == 1) // no
				bStripFilename = false;
			else
				// cancel
				return false;
			return true;

		}
		
		// ask user if wants to deal with unit
		private static boolean shouldDealWithUnit(String message, String title, DealWithUnit dealWithUnit) {
			if (!dealWithUnit.askAgain) {
				return dealWithUnit.result;
			}
			JPanel messagePanel = new JPanel();
			messagePanel.setLayout(new GridLayout(2,1));
			JCheckBox noAskAgainCB = new JCheckBox("Do not show again");
			JLabel messageLbl = new JLabel(message);
			messagePanel.add(messageLbl);
			messagePanel.add(noAskAgainCB);
			int answer = JOptionPane.showOptionDialog(null, messagePanel, title, 
					JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, 
					null, null, null);
			if (answer == 0) // yes
				dealWithUnit.result = true;
			else if (answer == 1) // no
				dealWithUnit.result = false;
			
			dealWithUnit.askAgain = !noAskAgainCB.isSelected();
			
			return dealWithUnit.result;
		}
		
		// obfuscates user code
		private static boolean StripBodies(IRPModelElement modelElement) {

			try {
				if (modelElement instanceof IRPOperation) {
					IRPOperation rpOperation = (IRPOperation) modelElement;
					if (shouldObfuscateField("OperationBody")){
						String newValue = getFieldContent("OperationBody");
						if (rpOperation.getIsCtor() == 0 && 
								rpOperation.getBody() != null && rpOperation.getBody().length() > 0) {
							rpOperation.setBody(newValue);
						}
						if (rpOperation.getIsCtor() == 1 &&
								rpOperation.getInitializer() != null && rpOperation.getInitializer().length() > 0) {
							rpOperation
									.setInitializer(newValue);
						}
					}
				}
				if (modelElement instanceof IRPState) {
					IRPState state = (IRPState) modelElement;
					if (shouldObfuscateField("StateEntryAction")){
						String newValue = getFieldContent("StateEntryAction");
						if (state.getEntryAction() != null && state.getEntryAction().length() > 0) {
							state.setEntryAction(newValue);
						}
					}
					if (shouldObfuscateField("StateExitAction")){
						String newValue = getFieldContent("StateExitAction");
						if (state.getExitAction() != null && state.getExitAction().length() > 0) {
							state.setExitAction(newValue);
						}
					}
				}
				if (modelElement instanceof IRPTransition) {
					IRPTransition transition = (IRPTransition) modelElement;
					IRPAction action = transition.getItsAction();
					if (action != null && action.getBody().length() > 0)
						if (shouldObfuscateField("TransitionAction")){
							String newValue = getFieldContent("TransitionAction");
							transition
									.setItsAction(newValue);
						}
					IRPGuard guard = transition.getItsGuard();
					if (guard != null && guard.getBody().length() > 0)
						if (shouldObfuscateField("TransitionGuard")){
							String newValue = getFieldContent("TransitionGuard");
							transition
									.setItsGuard(newValue);
						}
				}
				if (modelElement instanceof IRPType) {
					IRPType type = (IRPType) modelElement;
					if (type.getKind().equals("Language"))
						if (shouldObfuscateField("TypeDeclaration")){
							String newValue = getFieldContent("TypeDeclaration");
							if (type.getDeclaration() != null && type.getDeclaration().length() > 0) {
								type
										.setDeclaration(newValue);
							}
						}
				}
				if (modelElement instanceof IRPAnnotation) {
					IRPAnnotation annotation = (IRPAnnotation) modelElement;
					if (shouldObfuscateField("AnnotationAndComment")){
						String newValue = getFieldContent("AnnotationAndComment");
						if (annotation.getSpecification() != null && annotation.getSpecification().length() > 0) {
							annotation
									.setSpecification(newValue);
						}
					}
				}
				if (modelElement instanceof IRPVariable) {
					IRPVariable variable = (IRPVariable) modelElement;
					if (shouldObfuscateField("VariableDefaultValue")){
						String newValue = getFieldContent("VariableDefaultValue");
						if (variable.getDefaultValue() != null && variable.getDefaultValue().length() > 0) {
							variable
									.setDefaultValue(newValue);
						}
					}

				}

			} catch (RhapsodyRuntimeException e) {
				String errorMessage = m_rhpApplication.errorMessage();
				if (modelElement instanceof IRPState) {
					if (((IRPState) modelElement).getInheritsFrom() != null)
						return true; // will be solved later in parent state
				}
				printError(errorMessage, modelElement);
				return false;
			}
			return true;

		}

		// obfuscate description
		private static boolean StripDescription(IRPModelElement modelElement) {

			try {
				boolean bShouldStrip = true;

				if (modelElement instanceof IRPType) {
					IRPType type = (IRPType) modelElement;
					if (type.getIsPredefined() == 1)
						bShouldStrip = false;
				}

				if (bShouldStrip)
					if (shouldObfuscateField("ElementDescription")){
						String newValue = getFieldContent("ElementDescription");
						if (modelElement.getDescription() != null && modelElement.getDescription().length() > 0) {
							String htmlDescription = modelElement.getDescriptionHTML();
							Pattern pat = Pattern.compile("<a href=\\\".*\\\">");
							Matcher mat = pat.matcher(htmlDescription);
							String links = "";
							while (mat.find()) {
								links += "\n";
								links += mat.group();
							}
							modelElement.setDescription(newValue + (links.length() > 0 ? ("\n\nLinks:" + links) : ""));
						}
						
					}

			} catch (RhapsodyRuntimeException e) {
				printError(m_rhpApplication.errorMessage(), modelElement);
				return false;
			}
			return true;

		}

		// obfuscates model element - recursive.
		public static void StripElements(IRPModelElement parentElement) {
			
			ProgressUpdater progressUpdater = null;
			
			if (parentElement.equals(rootElement)) {
				if (rootElement instanceof IRPUnit) {
					IRPUnit rootUnit = (IRPUnit)rootElement;
					if (rootUnit.isReadOnly() == 1 && !UnitLoadManager.getInstance().isReferenceUnit(rootUnit)) {
						if (!shouldDealWithUnit(
								"<html>Unit " + rootUnit.getName() + " is " +  
								"Read Only" +
								"<BR> Include it in obfuscated model ?" +
								"<BR>   (Will make it Read-Write for the obfuscation process)</html>", 
								"Unit is Read Only",
								readOnly)
								)
								 {
							printError("Read-Only Unit " + rootUnit.getName()
									+ " Was Skipped.", rootUnit);
							return;
	
						} else {
							rootUnit.setReadOnly(0);
						}
					}
				}
				if (rootElement instanceof IRPProject) {
					rootProject = (IRPProject)parentElement;
				} else {
					rootProject = rootElement.getProject();
				}
				unitsNotToObfuscate = showDialogSelectUnitsNotToObfuscate(parentElement.getSaveUnit());
				propertiesToObfuscate = getObfuscatorProperty("PropertiesToObfuscate");
				progressUpdater = new ProgressUpdater();
				progressUpdater.bringToFront();
			}
			
			final String className = "TopLevel";
			final String metaClass = "Class";
			if (parentElement instanceof IRPUnit) {
				// m_rhpApplication.writeToOutputWindow("","Obfuscating " +
				// parentElement.getFullPathNameIn() + "\n");
				System.out.println("Obfuscating "
						+ parentElement.getFullPathNameIn());
			}

			IRPCollection coll = null;
			coll = parentElement.getNestedElements();
			int numOfElements = coll.getCount(); 
			List<String> elementsGUIDs = new ArrayList<String>();
			for (int i = 1; i <= numOfElements; i++) {
				elementsGUIDs.add(((IRPModelElement)coll.getItem(i)).getGUID());
			}
			IRPProject parentElementProject = parentElement.getProject();
			for (int i = 0; i < numOfElements; i++) {
				IRPModelElement modelElement = parentElementProject.findElementByGUID(elementsGUIDs.get(i));
				if (progressUpdater != null) {
					progressUpdater.setProgress(i*100/numOfElements);
				}
				
				if (isPreDefinedPackage(modelElement)) {
					continue;
				}
				
				boolean bShouldStrip = true;
				int reloadMode = -1;

				if (modelElement instanceof IRPUnit) {
					IRPUnit mUnit = (IRPUnit) modelElement;
					if (mUnit.getIsExternal() == 1) {
						printError("Element used as external " + modelElement.getName()
								+ " Was Skipped.", modelElement);
						continue; // no need to strip the element or it nested
					}
					
					String unitDir = null;
					IRPProject currProject = null;
					String projectDir = null;
					File unitFileCurrentLocation = null;
					if (mUnit.isSeparateSaveUnit() == 1) {
						unitDir = mUnit.getCurrentDirectory();
						currProject = mUnit.getProject();
						projectDir = currProject.getCurrentDirectory();
						unitFileCurrentLocation = new File(m_rhpApplication.expandStringKeywords(unitDir + 
								System.getProperty("file.separator") +
								mUnit.getFilename()));
						if (!unitFileCurrentLocation.exists() || !unitFileCurrentLocation.isFile()) {
							printError("File \n\t\t " + unitFileCurrentLocation.getAbsolutePath() + "\nfor " + modelElement.getUserDefinedMetaClass() + " " + modelElement.getName()
									+ "\n\twas not found. \nElement (and it's content) was skipped.", modelElement);
							continue;
						}
					}
					
					// elements
					if (mUnit.getIsStub() == 1) {
						if (!shouldDealWithUnit(
								"<html>Unit " + mUnit.getName() + " is " +  
								"not loaded." +
								"<BR> Include it in obfuscated model ?" +
								"<BR>   (This will load it into the model)</html>", 
								"Unit is unloaded",
								unloaded)
								)
								 {
							bShouldStrip = false;
							printError("Unloaded Unit " + modelElement.getName()
									+ " Was Skipped.", modelElement);

						} else {
							bShouldStrip = true;
							String unitGUID = mUnit.getGUID();
							IRPProject currProj = mUnit.getProject();
							mUnit.load(1);
							modelElement = currProj.findElementByGUID(unitGUID);
							if (modelElement instanceof IRPUnit) {
								mUnit = (IRPUnit)modelElement;
							}
						}
					}
					if (mUnit.isReadOnly() == 1 && !UnitLoadManager.getInstance().isReferenceUnit(mUnit)) {
						if (!shouldDealWithUnit(
								"<html>Unit " + mUnit.getName() + " is " +  
								"Read Only" +
								"<BR> Include it in obfuscated model ?" +
								"<BR>   (Will make it Read-Write for the obfuscation process)</html>", 
								"Unit is Read Only",
								readOnly)
								)
								 {
							bShouldStrip = false;
							printError("Read-Only Unit " + modelElement.getName()
									+ " Was Skipped.", modelElement);

						} else {
							bShouldStrip = true;
							mUnit.setReadOnly(0);
						}
					}
					
					if (mUnit.isSeparateSaveUnit() == 1 && 
							UnitLoadManager.getInstance().isNOTAsUnitWithCopy(mUnit)) {
						if (!shouldDealWithUnit(
								"<html>Unit " + mUnit.getName() + " is " +  
								((UnitLoadManager.getInstance().isReferenceUnit(mUnit)) ? "Referenced Unit" : "Not Copied Into Model") +
								"<BR> Include it in obfuscated model ?</html>", 
								"Unit is " + ((UnitLoadManager.getInstance().isReferenceUnit(mUnit)) ? "Referenced Unit" : "Not Copied Into Model"),
								((UnitLoadManager.getInstance().isReferenceUnit(mUnit)) ? reference :  notCopied)
								)
								) {
							continue;
						}

						File unitFileNewLocation = new File(projectDir + System.getProperty("file.separator") + UnitLoadManager.OBFUSCATED_MODEL +  System.getProperty("file.separator") + 
								((UnitLoadManager.getInstance().isReferenceUnit(mUnit)) ? UnitLoadManager.REFERENCED_UNITS : UnitLoadManager.UNITS_NOT_COPIED_INTO_MODEL) + 
								System.getProperty("file.separator") +
								mUnit.getFilename());
						if (unitFileNewLocation.getParentFile().exists() && unitFileNewLocation.getParentFile().isFile()) {
							throw new RuntimeException("Unable to create folder for obfuscated referenced units");
						} else {
							unitFileNewLocation.getParentFile().mkdirs();
						}
						UnitLoadManager.getInstance().fileCopy(mUnit.getGUID(), unitFileCurrentLocation, unitFileNewLocation, true);
						String elementGUID = mUnit.getGUID();
						IRPModelElement unitOwner = mUnit.getOwner();
						reloadMode = UnitLoadManager.getInstance().getUnitLoadMode(mUnit);
						mUnit.unload();
						UnitLoadManager.getInstance().loadUnitByLoadType(m_rhpApplication, mUnit, unitFileNewLocation.getAbsolutePath(), UnitLoadManager.LoadMode.AS_UNIT_WITHOUT_COPY, true);
						modelElement = currProject.findElementByGUID(elementGUID);
						mUnit = (IRPUnit)modelElement;
						if (!mUnit.getOwner().equals(unitOwner)) {
							mUnit.setOwner(unitOwner);
						}
						unitsToSetPath.add(elementGUID);
						if (! unitsNotToObfuscate.contains(elementGUID)){ 
							/// will be saved to the same name - so we do not want to delete it
							unitsToDelete.add(unitFileNewLocation.getAbsolutePath());
						}
					}
				}

				// strip nested elements including package top level class
				StripElements(modelElement); // strip the nested elements
				if (bShouldStrip) {
					if (modelElement instanceof IRPPackage) {
						IRPPackage mPackage = (IRPPackage) modelElement;
						IRPModelElement topLevelClass = mPackage
								.findNestedElement(className, metaClass);
						if (topLevelClass != null)
							StripElements(topLevelClass);
					}
				}

				// check if elements need to obfuscate
				if (modelElement instanceof IRPClass) {
					IRPModelElement owner = modelElement.getOwner();
					if (modelElement instanceof IRPRelation) {
						if (((IRPRelation) owner).isTypelessObject() == 1) {
							bShouldStrip = false;
						}
					}
					if (owner instanceof IRPInstance) {
						bShouldStrip = false;
					}
					if (modelElement.getName().equals("TopLevel")) {
						if (owner instanceof IRPPackage)
							bShouldStrip = false;
					}
				}
				if (bShouldStrip && modelElement instanceof IRPEventReception) {
					bShouldStrip = false;
				}
				
				if (unitsNotToObfuscate.contains(modelElement.getSaveUnit().getGUID())){
					System.out.println(modelElement.getFullPathName() + " is in unit selected as not obfuscated");
					bShouldStrip = false;
				}
				
				// strip the element
				if (bShouldStrip) {
					boolean bSucsess = true;
					bSucsess &= StripNames(modelElement);
					bSucsess &= StripBodies(modelElement);
					bSucsess &= StripDescription(modelElement);
					bSucsess &= StripNotes(modelElement);
					bSucsess &= StripPropertyValues(modelElement);
					if (bSucsess)
						addToStrippedElements(modelElement);
				}
				if (reloadMode != -1) {
					if (modelElement instanceof IRPUnit) {
						IRPUnit mUnit = (IRPUnit) modelElement;
						IRPProject currProject = mUnit.getProject();
						String unitDir = mUnit.getCurrentDirectory();
						File referencedFileCurrentLocation = new File(m_rhpApplication.expandStringKeywords(unitDir + 
								System.getProperty("file.separator") +
								mUnit.getFilename()));
						String elementGUID = mUnit.getGUID();
						IRPModelElement unitOwner = mUnit.getOwner();
						UnitLoadManager.getInstance().setUnitPath(mUnit,currProject.getCurrentDirectory() +
								System.getProperty("file.separator") + UnitLoadManager.OBFUSCATED_MODEL +  System.getProperty("file.separator") + 
								((UnitLoadManager.LoadMode.AS_REFERENCE == reloadMode) ? UnitLoadManager.REFERENCED_UNITS : UnitLoadManager.UNITS_NOT_COPIED_INTO_MODEL) );
						mUnit.save(0);
						String unitFileName = mUnit.getFilename();
						referencedFileCurrentLocation = new File(currProject.getCurrentDirectory() + 
								System.getProperty("file.separator") + UnitLoadManager.OBFUSCATED_MODEL + System.getProperty("file.separator") + 
								((UnitLoadManager.LoadMode.AS_REFERENCE == reloadMode) ? UnitLoadManager.REFERENCED_UNITS : UnitLoadManager.UNITS_NOT_COPIED_INTO_MODEL)
								 +  System.getProperty("file.separator")
								+ unitFileName);
						mUnit.unload();
						UnitLoadManager.getInstance().loadUnitByLoadType(m_rhpApplication, mUnit, referencedFileCurrentLocation.getAbsolutePath(), reloadMode, false);
						modelElement = currProject.findElementByGUID(elementGUID);
						mUnit = (IRPUnit)modelElement;
						UnitLoadManager.getInstance().setUnitPath(mUnit,currProject.getCurrentDirectory() +
								System.getProperty("file.separator") + UnitLoadManager.OBFUSCATED_MODEL +  System.getProperty("file.separator") + 
								((UnitLoadManager.LoadMode.AS_REFERENCE == reloadMode) ? UnitLoadManager.REFERENCED_UNITS : UnitLoadManager.UNITS_NOT_COPIED_INTO_MODEL) );
						if (!mUnit.getOwner().equals(unitOwner)) {
							mUnit.setOwner(unitOwner);
						}
						UnitLoadManager.getInstance().setUnitPath(mUnit, (UnitLoadManager.getInstance().getUnitLoadMode(unitOwner.getSaveUnit()) == reloadMode) ? "" :
							  (".." + 
								System.getProperty("file.separator") + 
								((UnitLoadManager.LoadMode.AS_REFERENCE == reloadMode) ? UnitLoadManager.REFERENCED_UNITS : UnitLoadManager.UNITS_NOT_COPIED_INTO_MODEL)));
					}
				}
			}
			if (progressUpdater != null) {
				progressUpdater.close();
			}
		}

		private static boolean StripPropertyValues(IRPModelElement modelElement) {
			try {
				IRPCollection overriddenProperties = modelElement.getOverriddenProperties(0);
				int overridenPropertiesCount = overriddenProperties.getCount();
				for (int opIdx = 1; opIdx <= overridenPropertiesCount; opIdx++) {
					String overridenProp = (String)overriddenProperties.getItem(opIdx);
					String propVal = overridenProp.substring(overridenProp.indexOf(':')+1);
					String propName = overridenProp.substring(0,overridenProp.indexOf(':'));
					if (propertiesToObfuscate.contains(propName)) {
						String[] splittedPropVal = propVal.split(",");
						String obfuscatedPropVal = "";
						for (int splittedValIdx = 0; splittedValIdx < splittedPropVal.length; splittedValIdx++) {
							obfuscatedPropVal += GetNewName(splittedPropVal[splittedValIdx].trim()) + 
									(((splittedValIdx + 1) < splittedPropVal.length) ? ", " : ""); 
						}
						modelElement.setPropertyValue(propName, obfuscatedPropVal);
					}
				}
			} catch (Exception e){
				System.out.println("error obfuscating properties for " + modelElement.getFullPathName());
				return false;
			}
			return true;
		}
		
		private static List<String> getNestedSaveUnitsRecursive(IRPUnit root) {
			List<String> retList = new ArrayList<String>();
			IRPCollection nestedSaveUnits = root.getNestedSaveUnits();
			int nestedUnitsCount = nestedSaveUnits.getCount();
			for (int i = 1; i <= nestedUnitsCount; i++) {
				IRPModelElement modelElement = (IRPModelElement)nestedSaveUnits.getItem(i);
				retList.add(modelElement.getGUID());
				retList.addAll(getNestedSaveUnitsRecursive((IRPUnit)modelElement));
			}
			return retList;
		}
		
		private static String getFieldContent(String fieldToGet) {
			List<String> fieldValueList = getObfuscatorProperty(fieldToGet + "Content");
			if (fieldValueList == null || fieldValueList.size() == 0) {
				return "Obfuscated " + fieldToGet;
			}
			return fieldValueList.get(0);
		}

		private static boolean shouldObfuscateField(String fieldToGet) {
			List<String> fieldValueList = getObfuscatorProperty("Remove" + fieldToGet + "Content");
			if (fieldValueList == null || fieldValueList.size() == 0) {
				return true;
			}
			return !fieldValueList.get(0).equalsIgnoreCase("False");
		}

		
		private static List<String> getObfuscatorProperty(String property) {
			List<String> retList = new ArrayList<String>();
			String valueString = getObfuscatorProperties().getProperty(property);
			if (valueString == null) {
				return retList;
			}
			String[] valueStringArray = valueString.split("\n");
			for (int i = 0; i < valueStringArray.length; i++) {
				valueStringArray[i] = valueStringArray[i].trim();
			}
			retList.addAll(Arrays.asList(valueStringArray));
			return retList;
		}

		private static Properties getObfuscatorProperties() {
			if (obfuscatorProperties != null) {
				return obfuscatorProperties;
			}
			URL obfuscatorIniFile = RhpObfuscator.class.getClassLoader().getResource("com/ibm/rhapsody/samples/obfuscator/obfuscator.ini");
			if (obfuscatorIniFile != null) {
				BufferedReader bReader = null;
				try {
					bReader = new BufferedReader(new InputStreamReader(new FileInputStream(new File(obfuscatorIniFile.toURI()))));
					obfuscatorProperties = new Properties();
					obfuscatorProperties.load(bReader);
					return obfuscatorProperties;
				} catch (FileNotFoundException | URISyntaxException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} finally {
					if (bReader != null) {
						try {
							bReader.close();
						} catch (IOException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
					}
				}
			}
			return null;
		}

		private static List<String> showDialogSelectUnitsNotToObfuscate(IRPUnit unit) {
			List<String> predefinedUnitsToExcludeFromObfuscation = getObfuscatorProperty("UnitsToExclude");
			predefinedUnitsToExcludeFromObfuscation.addAll(predefinedTypesPackages);
			List<String> nestedUnitsGUIDs = getNestedSaveUnitsRecursive(unit);
			List<JCheckBox> nestedUnitsSelect = new ArrayList<JCheckBox>();
			JPanel panel = new JPanel();
			panel.setLayout(new GridLayout(nestedUnitsGUIDs.size()+1, 1));
			panel.add(new JLabel("Select units to obfuscate:"));
			JScrollPane scrollPane = new JScrollPane(panel, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
			scrollPane.setMaximumSize(new Dimension(500, (nestedUnitsGUIDs.size() > 8) ? 350 : (nestedUnitsGUIDs.size()+1)*35));
			scrollPane.setPreferredSize(new Dimension(500,(nestedUnitsGUIDs.size() > 8) ? 350 : (nestedUnitsGUIDs.size()+1)*35));
			for (int i = 0; i < nestedUnitsGUIDs.size(); i++) {
				IRPModelElement modelElement = (IRPModelElement)Obfuscator.rootProject.findElementByGUID(nestedUnitsGUIDs.get(i));
				if (modelElement instanceof IRPUnit) {
					JCheckBox checkBox = new JCheckBox(modelElement.getName() + " (" + modelElement.getFullPathName() + ")");
					nestedUnitsSelect.add(checkBox);
					panel.add(checkBox);
					checkBox.setSelected(!predefinedUnitsToExcludeFromObfuscation.contains(modelElement.getName()));
				}
			}
			JOptionPane.showMessageDialog(null, scrollPane, "Select elements to obfuscate", 
					JOptionPane.QUESTION_MESSAGE);
			List<String> retList = new ArrayList<String>();
			for (int i = 0 ; i < nestedUnitsSelect.size(); i++) {
				if (!nestedUnitsSelect.get(i).isSelected()) {
					retList.add(nestedUnitsGUIDs.get(i));
				}
			}
			return retList;
		}

		
		private static boolean isPreDefinedPackage(IRPModelElement modelElement) {
			if (modelElement.getOwner() instanceof IRPProject) {
				if (predefinedTypesPackages .contains(modelElement.getName().toLowerCase())) {
					return true;
				}
			}
			return false;
		}


		// obfuscates model element's name
		private static boolean StripNames(IRPModelElement modelElement) {

			try {

				// check if element needs to strip
				boolean bShouldStrip = true;
				if (modelElement instanceof IRPType) {
					IRPType type = (IRPType) modelElement;
					if (type.getIsPredefined() == 1) {
						bShouldStrip = false;
					}
					if (type.getName().equals("")) {
						bShouldStrip = false;
					}

				}

				if (bShouldStrip) {
					String strMetaClassName = modelElement.getMetaClass();
					if (strMetaClassName.equals("Project") ||
							strMetaClassName.equals("ActivityDiagram")) {
						bShouldStrip = false;
					}
					if (strMetaClassName.equals("State")
							&& modelElement.getName().equals("ROOT")) {
						bShouldStrip = false;
					}
				}
				
				if (bShouldStrip && modelElement instanceof IRPOperation) {
					IRPOperation rpOperation = (IRPOperation) modelElement;
					if (rpOperation.getIsCtor() == 1
							|| rpOperation.getIsDtor() == 1)
						bShouldStrip = false;

				}

				if (bShouldStrip && modelElement instanceof IRPRelation) {
					IRPRelation rpRelation = (IRPRelation) modelElement;
					rpRelation.setQualifier(GetNewName(rpRelation
							.getQualifier()));
				}

				if (bShouldStrip) {
					String strOrigUnitFilename = "";
					IRPUnit mUnit = null;
					if (modelElement instanceof IRPUnit) {
						mUnit = (IRPUnit) modelElement;
						if (mUnit.isSeparateSaveUnit() == 1) {
							strOrigUnitFilename = mUnit.getFilename();
							strOrigUnitFilename = strOrigUnitFilename
									.substring(0, strOrigUnitFilename
											.lastIndexOf("."));
						}
					}
					String strNewName = GetNewName(modelElement.getName());
					String strNewLabel = GetNewName(modelElement
							.getDisplayName());
					try {
						modelElement.setName(strNewName);
					} catch (RhapsodyRuntimeException e) {
						String errorMessage = m_rhpApplication.errorMessage();
						try {
							ModelElementDuplicateNameProcessor.getInstance().setModelElementName(modelElement, strNewName);
						} catch (RhapsodyRuntimeException nameSetSubError) {
							printError(errorMessage
									+ " (Maybe obfuscated later from base element)",
									modelElement);
						}
					}
					
					modelElement.setDisplayName(strNewLabel);
					
					if (mUnit != null && mUnit.isSeparateSaveUnit() == 1) {
						// force return to original name if bStripFilename is
						// false
						mUnit.setFilename(GetNewName(strOrigUnitFilename));
						if (bStripFilename)
							mUnit.setFilename(strNewName);
						else
							// force change
							mUnit.setFilename(strOrigUnitFilename);
					}
				}

			} catch (RhapsodyRuntimeException e) {
				String errorMessage = m_rhpApplication.errorMessage();
				if (modelElement instanceof IRPState) {
					if (((IRPState) modelElement).getInheritsFrom() != null)
						return true; // will be solved later in parent state
				}
				printError(errorMessage
						+ " (Maybe obfuscated later from base element)",
						modelElement);
				return false;
			}
			return true;
		}

		// obfuscate Notes
		private static boolean StripNotes(IRPModelElement modelElement) {

			try {
				if (modelElement instanceof IRPDiagram) {
					IRPDiagram d = (IRPDiagram) modelElement;
					IRPCollection col = d.getGraphicalElements();
					int numOfElements = col.getCount();
					for (int i = 1; i <= numOfElements; i++) {
						IRPGraphElement ge = (IRPGraphElement) col.getItem(i);
						IRPModelElement gemodel = ge.getModelObject();
						IRPGraphicalProperty gp = ge
								.getGraphicalProperty("Type");
						if (gp != null) {
							if (gp.getValue().equals("Note")
									|| gp.getValue().equals("FreeText")){
									if (shouldObfuscateField("GraphicNotesText")){
										String newValue = getFieldContent("GraphicNotesText");
										try {
											IRPGraphicalProperty textProp = ge.getGraphicalProperty("Text");
											if (textProp != null && textProp.getValue().length() > 0) {
												ge.setGraphicalProperty("Text",
														newValue);
											}
										} catch (RhapsodyRuntimeException e) {
											//// nothing to do here
										}
									}
							}else {
								if ((gemodel == null)
										|| (gemodel instanceof IRPMessage)) {
									gp = ge.getGraphicalProperty("Text");
									if (gp != null) {
										String textvalue = gp.getValue();
										ge.setGraphicalProperty("Text",
												GetNewName(textvalue));
									}
								}
							}
						}
						if (ge != null && ge instanceof IRPMessage) {
							gp = ge.getGraphicalProperty("FreeText");
							if (gp != null) {
								String textvalue = gp.getValue();
								ge.setGraphicalProperty("FreeText",
										GetNewName(textvalue));
							}
						}
					}
					if (shouldObfuscateField("ElementDescription")){
						String newValue = getFieldContent("ElementDescription");
						modelElement
								.setDescription(newValue);
					}
				}

			} catch (RhapsodyRuntimeException e) {
				printError(m_rhpApplication.errorMessage(), modelElement);
				return false;
			}
			return true;
		}
		
		private static void removeFilesWithUnobfuscatedNames() {
			if (bStripFilename) {
				for (int i = 0; i < unitsToDelete.size(); i++) {
					File fileToDelete = new File (unitsToDelete.get(i));
					if (fileToDelete.exists()) {
						fileToDelete.delete();
					}
				}
			}
		}

		public static void setPathRelativeToObfuscated(IRPProject project) {
			for (int i = 0; i < unitsToSetPath.size(); i++) {
				IRPModelElement modelElement = project.findElementByGUID(unitsToSetPath.get(i));
				if (modelElement instanceof IRPUnit) {
					IRPUnit currUnit = (IRPUnit)modelElement;
					String currUnitDir = currUnit.getCurrentDirectory();
					while (currUnitDir.endsWith("/") || currUnitDir.endsWith("\\")) {
						currUnitDir = currUnitDir.substring(0, currUnitDir.length() -1);
					}
					if (currUnitDir.endsWith(UnitLoadManager.UNITS_NOT_COPIED_INTO_MODEL)) {
						UnitLoadManager.getInstance().setUnitPath(currUnit, ".."+ System.getProperty("file.separator") +UnitLoadManager.UNITS_NOT_COPIED_INTO_MODEL);
					}
					if (currUnitDir.endsWith(UnitLoadManager.REFERENCED_UNITS)) {
						UnitLoadManager.getInstance().setUnitPath(currUnit, ".."+ System.getProperty("file.separator") +UnitLoadManager.UNITS_NOT_COPIED_INTO_MODEL);
					}
				}
			}
		}

		public static void setRootElement(IRPModelElement rootElement) {
			Obfuscator.rootElement = rootElement;
			
		}

	} // end of nested class Obfuscator

	// send error messages to java log file
	private static final boolean bLogErrorMessage = true;

	private static IRPApplication m_rhpApplication = null;

	private static Map<String, String> nonStrippedElementsGUID2FullName;
	private static Map<String, String> strippedElementsGUID2FullName;

	private static int nWarnings = 0;

	private static boolean verbose = false;

	private static boolean addNonStrippedElement(IRPModelElement modelelement) {
		if (verbose)
			System.out.println("addNonStrippedElement("
					+ modelelement.getName() + " " + modelelement.getGUID()
					+ ")");
		return (GetNonStrippedElementsGUID2FullName().put(
				modelelement.getGUID(),
				modelelement.getMetaClass() + " "
						+ modelelement.getFullPathNameIn()) == null);
	}

	private static String getNonStrippedElement(String GUID) {
		return GetNonStrippedElementsGUID2FullName().get(GUID);
	}

	private static String getNonStrippedElementsAsList() {
		String str = "";
		int i = 1;
		Set<String> keyset = GetNonStrippedElementsGUID2FullName().keySet();
		for (String guid : keyset) {
			if (getStrippedElement(guid) == null) {
				str += "\n" + i + ") " + getNonStrippedElement(guid);
				i++;
			}

		}
		return str;
	}

	private static boolean addToStrippedElements(IRPModelElement modelelement) {
		if (verbose)
			System.out.println("addToStrippedElements("
					+ modelelement.getName() + " " + modelelement.getGUID()
					+ ")");
		return (GetStrippedElementsGUID2FullName().put(
				modelelement.getGUID(),
				modelelement.getMetaClass() + " "
						+ modelelement.getFullPathNameIn()) == null);
	}

	private static String getStrippedElement(String GUID) {
		return GetStrippedElementsGUID2FullName().get(GUID);
	}

	private static Map<String, String> GetStrippedElementsGUID2FullName() {
		if (strippedElementsGUID2FullName == null)
			strippedElementsGUID2FullName = new HashMap<String, String>();

		return strippedElementsGUID2FullName;
	}

	private static Map<String, String> GetNonStrippedElementsGUID2FullName() {
		if (nonStrippedElementsGUID2FullName == null)
			nonStrippedElementsGUID2FullName = new HashMap<String, String>();

		return nonStrippedElementsGUID2FullName;
	}

	public static void main(String[] args) {

		try {
			// get the active Rhapsody application
			m_rhpApplication = RhapsodyAppServer.getActiveRhapsodyApplication();
			if (m_rhpApplication != null) {
				// get the active project
				IRPProject m_rpProject = m_rhpApplication.activeProject();
				if (m_rpProject != null) {
					if (Obfuscator.initShouldObfuscateUnitFileNames() == false)
						return;
					m_rhpApplication.writeToOutputWindow("",
							"Start obfuscating...");
					// start obfuscating from the project's root node
					String projectOrigName = m_rpProject.getName();
					Obfuscator.setRootElement(m_rpProject);
					Obfuscator.StripElements(m_rpProject);
					Obfuscator.removeFilesWithUnobfuscatedNames();
					m_rpProject.saveAs(m_rpProject.getCurrentDirectory() + System.getProperty("file.separator") + UnitLoadManager.OBFUSCATED_MODEL
							+ System.getProperty("file.separator") + (Obfuscator.bStripFilename ? Obfuscator.GetNewName(m_rpProject.getName()) : projectOrigName));
					m_rhpApplication.writeToOutputWindow("",
							"\nFinished obfuscating...");

					writeReport();
				}
			}
		} catch (RhapsodyRuntimeException e) {
			if (m_rhpApplication != null)
				m_rhpApplication.writeToOutputWindow("", m_rhpApplication
						.errorMessage());
			else
				JOptionPane.showMessageDialog(null, e.getMessage());
			e.printStackTrace();
		}
	}

	private static void writeReport() {
		String warningList = "";
		warningList = getNonStrippedElementsAsList();
		if (m_rhpApplication != null
				&& (warningList.equalsIgnoreCase("") == false)) {
			m_rhpApplication.writeToOutputWindow("",
					"\nElements failed to Obfuscate (read only etc.):");
			m_rhpApplication.writeToOutputWindow("",
					getNonStrippedElementsAsList());
		}

	}

	// prints an error message to Rhapsody output and/or Java log file
	public static void printError(String errorMessage, IRPModelElement element) {
		errorMessage = "Obfuscating " + element.getMetaClass() + " : "
				+ element.getFullPathNameIn() + "\n" + errorMessage;

		System.out.println(errorMessage);

		if (bLogErrorMessage)
			RhapsodyAppServer.writeToLog(errorMessage);

		nWarnings++;
		addNonStrippedElement(element);
	}

	// called when the plug-in pop-up menu (if applicable) is selected
	@Override
	public void OnMenuItemSelect(String menuItem) {
		try {
			// show the selected element name
			String selElemName = null;
			nonStrippedElementsGUID2FullName = null;
			IRPModelElement element = m_rhpApplication.getSelectedElement();
			if (element != null) {
				selElemName = element.getName();
			} else {
				m_rhpApplication.writeToOutputWindow("Obfuscator",
						"You have not selected any elements to obfuscate.");
				return;
			}
	
			if (Obfuscator.initShouldObfuscateUnitFileNames() == false)
				return;
	
			String msg = "Obfuscating " + selElemName + "...\n";
			m_rhpApplication.writeToOutputWindow("", msg);
			Obfuscator.setRootElement(element);
			Obfuscator.StripElements(element);
			m_rhpApplication.writeToOutputWindow("", "\nFinished Obfuscating");
			writeReport();
		} catch(Exception e) {
			String stackTrace;
			StringWriter eStringWriter = new StringWriter();
			PrintWriter ePrintWriter = new PrintWriter(eStringWriter);
			e.printStackTrace(ePrintWriter);
			stackTrace = eStringWriter.toString();
			m_rhpApplication.writeToOutputWindow("", stackTrace);
		}

	}

	@Override
	public void OnTrigger(String trigger) {
	}

	// called when the project is closed - if true is returned the plugin will
	// be unloaded
	@Override
	public boolean RhpPluginCleanup() {
		// cleanup
		m_rhpApplication = null;
		// return true so the plug-in will be unloaded now (on project close)
		return true;
	}

	// called when Rhapsody exits
	@Override
	public void RhpPluginFinalCleanup() {
	}

	// called when the plug-in is loaded
	@Override
	public void RhpPluginInit(final IRPApplication rpyApplication) {
		// keep the application interface for later use
		m_rhpApplication = rpyApplication;
		String msg = "Obfuscator Plugin was loaded.\n";
		m_rhpApplication.writeToOutputWindow("", msg);

	}

	// called when the plug-in menu item under the "Tools" menu is selected
	@Override
	public void RhpPluginInvokeItem() {
		// show the active project overridden properties
		if (m_rhpApplication != null) {
			try {
				IRPProject prj = m_rhpApplication.activeProject();
				if (prj != null) {
					m_rhpApplication
							.writeToOutputWindow("", "Start obfuscating...");
	
					if (Obfuscator.initShouldObfuscateUnitFileNames() == false)
						return;
					nonStrippedElementsGUID2FullName = null;
					String projectOrigName = prj.getName();
					Obfuscator.setRootElement(prj);
					Obfuscator.StripElements(prj);
					Obfuscator.removeFilesWithUnobfuscatedNames();
					prj.saveAs(prj.getCurrentDirectory() + System.getProperty("file.separator") + UnitLoadManager.OBFUSCATED_MODEL
							+ System.getProperty("file.separator") + (Obfuscator.bStripFilename ? Obfuscator.GetNewName(prj.getName()) : projectOrigName));
	
					m_rhpApplication.writeToOutputWindow("",
							"\nFinished obfuscating...");
	
					writeReport();
	
				} else {
					m_rhpApplication.writeToOutputWindow("",
							"There is currently no active Rhapsody project.");
				}
			} catch (Exception e) {
				String stackTrace;
				StringWriter eStringWriter = new StringWriter();
				PrintWriter ePrintWriter = new PrintWriter(eStringWriter);
				e.printStackTrace(ePrintWriter);
				stackTrace = eStringWriter.toString();
				m_rhpApplication.writeToOutputWindow("", stackTrace);
			}
		}
	}
}
